# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
from typing import Dict, Any, List, Tuple
from abc import abstractmethod
import struct

from ezdxf.lldxf.const import acad_release_to_dxf_version
from ezdxf.tools.binarydata import BitStream

from .const import *
from .crc import crc8
from .fileheader import FileHeader


def load_header_section(specs: FileHeader, data: Bytes, crc_check=False):
    if specs.version <= ACAD_2000:
        return DwgHeaderSectionR2000(specs, data, crc_check)
    else:
        return DwgHeaderSectionR2004(specs, data, crc_check)


class DwgSectionLoader:
    def __init__(self, specs: FileHeader, data: Bytes, crc_check=False):
        self.specs = specs
        self.crc_check = crc_check
        self.data = self.load_data_section(data)

    @abstractmethod
    def load_data_section(self, data: Bytes) -> Bytes:
        ...


class DwgHeaderSectionR2000(DwgSectionLoader):
    def load_data_section(self, data: Bytes) -> Bytes:
        if self.specs.version > ACAD_2000:
            raise DwgVersionError(self.specs.version)
        seeker, section_size = self.specs.sections[HEADER_ID]
        return data[seeker:seeker + section_size]

    def load_header_vars(self) -> Dict:
        data = self.data
        sentinel = data[:16]
        if sentinel != b'\xCF\x7B\x1F\x23\xFD\xDE\x38\xA9\x5F\x7C\x68\xB8\x4E\x6D\x33\x5F':
            raise DwgCorruptedHeaderSection('Sentinel for start of HEADER section not found.')
        index = 16
        size = struct.unpack_from('<L', data, index)[0]
        index += 4
        bs = BitStream(data[index: index + size], dxfversion=self.specs.version, encoding=self.specs.encoding)
        hdr_vars = parse_header(bs)
        index += size
        if self.crc_check:
            check = struct.unpack_from('<H', data, index)[0]
            # CRC of data from end of sentinel until start of crc value
            crc = crc8(data[16:-18], seed=0xc0c1)
            if check != crc:
                raise CRCError('CRC error in header section.')
        sentinel = data[-16:]
        if sentinel != b'\x30\x84\xE0\xDC\x02\x21\xC7\x56\xA0\x83\x97\x47\xB1\x92\xCC\xA0':
            raise DwgCorruptedHeaderSection('Sentinel for end of HEADER section not found.')
        return hdr_vars


class DwgHeaderSectionR2004(DwgHeaderSectionR2000):
    def load_data(self, data: Bytes) -> Bytes:
        raise NotImplementedError()


CMD_SET_VERSION = 'ver'
CMD_SKIP_BITS = 'skip_bits'
CMD_SKIP_NEXT_IF = 'skip_next_if'
CMD_SET_VAR = 'var'


def _min_max_versions(version: str) -> Tuple[str, str]:
    min_ver = ACAD_13
    max_ver = ACAD_LATEST
    if version != 'all':
        v = version.split('-')
        if len(v) > 1:
            min_ver = acad_release_to_dxf_version[v[0].strip()]
            max_ver = acad_release_to_dxf_version[v[1].strip()]
        else:
            v = v[0].strip()
            if v[-1] == '+':
                min_ver = acad_release_to_dxf_version[v[:-1]]
            else:
                min_ver = max_ver = acad_release_to_dxf_version[v]
    return min_ver, max_ver


def load_commands(desc: str) -> List[Tuple[str, Any]]:
    commands = []
    lines = desc.split('\n')
    for line in lines:
        line = line.strip()
        if not line or line[0] == '#':
            continue
        try:
            command, param = line.split(':')
        except ValueError:
            raise ValueError(f'Unpack Error in line: {line}')
        command = command.strip()
        param = param.split('#')[0].strip()
        if command == CMD_SET_VERSION:
            commands.append((CMD_SET_VERSION, _min_max_versions(param)))
        elif command in {CMD_SKIP_BITS, CMD_SKIP_NEXT_IF}:
            commands.append((command, param))
        elif command[0] == '$':
            commands.append((CMD_SET_VAR, (command, param)))
        else:
            raise ValueError(f'Unknown command: {command}')
    return commands


def parse_bitstream(bs: BitStream, commands: List[Tuple[str, Any]]) -> Dict[str, Any]:
    version = bs.dxfversion
    min_ver = ACAD_13
    max_ver = ACAD_LATEST
    hdr_vars = dict()
    skip_next_cmd = False
    for cmd, params in commands:
        if skip_next_cmd:
            skip_next_cmd = False
            continue

        if cmd == CMD_SET_VERSION:
            min_ver, max_ver = params
        elif cmd == CMD_SKIP_BITS:
            bs.skip(int(params))
        elif cmd == CMD_SKIP_NEXT_IF:
            skip_next_cmd = eval(params, None, {'header': hdr_vars})
        elif cmd == CMD_SET_VAR:
            if min_ver <= version <= max_ver:
                name, code = params
                hdr_vars[name] = bs.read_code(code)
        else:
            raise ValueError(f'Unknown command: {cmd}')
    return hdr_vars


def parse_header(bs: BitStream) -> Dict[str, Any]:
    commands = load_commands(HEADER_DESCRIPTION)
    return parse_bitstream(bs, commands)


HEADER_DESCRIPTION = """
ver: R2007
$SIZE_IN_BITS: RL # Size in bits 

ver: R2013+ 
$REQUIREDVERSIONS: BLL # default value 0, read only 

ver: all
$UNKNOWN: BD # Unknown, default value 412148564080.0 
$UNKNOWN: BD # Unknown, default value 1.0 
$UNKNOWN: BD # Unknown, default value 1.0 
$UNKNOWN: BD # Unknown, default value 1.0 
$UNKNOWN: TV # Unknown text string, default "" 
$UNKNOWN: TV # Unknown text string, default "" 
$UNKNOWN: TV # Unknown text string, default "" 
$UNKNOWN: TV # Unknown text string, default "" 
$UNKNOWN: BL # Unknown long, default value 24L 
$UNKNOWN: BL # Unknown long, default value 0L; 

ver: R13-R14 
$UNKNOWN: BS # Unknown short, default value 0 

ver: R13-R2000 
$CURRENT_VIEWPORT_ENTITY_HEADER: H # Handle of the current viewport entity header (hard pointer) 

ver: all
$DIMASO: B 
$DIMSHO: B 

ver: R13-R14 
$DIMSAV: B # Undocumented 

ver: all
$PLINEGEN: B 
$ORTHOMODE: B 
$REGENMODE: B 
$FILLMODE: B 
$QTEXTMODE: B 
$PSLTSCALE: B 
$LIMCHECK: B 

ver: R13-R14 
$BLIPMODE: B 

ver: R2004+ 
$UNKNOWN: B #  Undocumented 

ver: all 
$USRTIMER: B #  (User timer on/off) 
$SKPOLY: B 
$ANGDIR: B 
$SPLFRAME: B 

ver: R13-R14 
$ATTREQ: B
$ATTDIA: B

ver: all 
$MIRRTEXT: B 
$WORLDVIEW: B

ver: R13-R14 
$WIREFRAME: B # Undocumented. 

ver: all
$TILEMODE: B 
$PLIMCHECK: B 
$VISRETAIN: B 

ver: R13-R14 
$DELOBJ: B

ver: all
$DISPSILH: B 
$PELLIPSE: B # (not present in DXF) 
$PROXYGRAPHICS: BS 

ver: R13-R14 
$DRAGMODE: BS 

ver: all 
$TREEDEPTH: BS 
$LUNITS: BS 
$LUPREC: BS 
$AUNITS: BS 
$AUPREC: BS 

ver: R13-R14 
$OSMODE: BS 

ver: all 
$ATTMODE: BS 

ver: R13-R14 
$COORDS: BS 

ver: all 
$PDMODE: BS 

ver: R13-R14 
$PICKSTYLE: BS 

ver: R2004+
$UNKNOWN: BL 
$UNKNOWN: BL 
$UNKNOWN: BL 

ver: all 
$USERI1: BS 
$USERI2: BS 
$USERI3: BS 
$USERI4: BS 
$USERI5: BS 
$SPLINESEGS: BS 
$SURFU: BS 
$SURFV: BS 
$SURFTYPE: BS 
$SURFTAB1: BS 
$SURFTAB2: BS 
$SPLINETYPE: BS 
$SHADEDGE: BS 
$SHADEDIF: BS 
$UNITMODE: BS 
$MAXACTVP: BS 
$ISOLINES: BS 
$CMLJUST: BS 
$TEXTQLTY: BS 
$LTSCALE: BD 
$TEXTSIZE: BD 
$TRACEWID: BD 
$SKETCHINC: BD 
$FILLETRAD: BD 
$THICKNESS: BD 
$ANGBASE: BD 
$PDSIZE: BD 
$PLINEWID: BD 
$USERR1: BD 
$USERR2: BD 
$USERR3: BD 
$USERR4: BD 
$USERR5: BD 
$CHAMFERA: BD 
$CHAMFERB: BD 
$CHAMFERC: BD 
$CHAMFERD: BD 
$FACETRES: BD 
$CMLSCALE: BD 
$CELTSCALE: BD 

ver: R13-R2004 
$MENUNAME: TV

ver: all 
$TDCREATE: BL # (Julian day) 
$TDCREATE: BL # (Milliseconds into the day) 
$TDUPDATE: BL # (Julian day) 
$TDUPDATE: BL # (Milliseconds into the day) 

ver: R2004+
$UNKNOWN: BL 
$UNKNOWN: BL 
$UNKNOWN: BL 

ver: all 
$TDINDWG: BL # (Days) 
$TDINDWG: BL # (Milliseconds into the day) 
$TDUSRTIMER: BL # (Days) 
$TDUSRTIMER: BL # (Milliseconds into the day) 
$CECOLOR: CMC

# with an 8-bit length specifier preceding the handle bytes (standard hex handle form) (code 0). 
# The HANDSEED is not part of the handle stream, but of the normal data stream (relevant for R21 and later). 

$HANDSEED: H # The next handle 
$CLAYER: H # (hard pointer) 
$TEXTSTYLE: H # (hard pointer) 
$CELTYPE: H # (hard pointer) 

ver: R2007+
$CMATERIAL: H # (hard pointer) 

ver: all 
$DIMSTYLE: H # (hard pointer) 
$CMLSTYLE: H # (hard pointer) 

ver: R2000+ 
$PSVPSCALE: BD 

ver: all 
$PINSBASE: 3BD # (PSPACE) 
$PEXTMIN: 3BD # (PSPACE) 
$PEXTMAX: 3BD # (PSPACE) 
$PLIMMIN: 2RD # (PSPACE) 
$PLIMMAX: 2RD # (PSPACE) 
$PELEVATION: BD # (PSPACE) 
$PUCSORG: 3BD # (PSPACE) 
$PUCSXDIR: 3BD # (PSPACE) 
$PUCSYDIR: 3BD # (PSPACE) 
$PUCSNAME: H # (PSPACE) (hard pointer) 

ver: R2000+
$PUCSORTHOREF: H # (hard pointer) 
$PUCSORTHOVIEW: BS 
$PUCSBASE: H # (hard pointer) 
$PUCSORGTOP: 3BD 
$PUCSORGBOTTOM: 3BD 
$PUCSORGLEFT: 3BD 
$PUCSORGRIGHT: 3BD 
$PUCSORGFRONT: 3BD 
$PUCSORGBACK: 3BD 

ver: all 
$INSBASE: 3BD # (MSPACE) 
$EXTMIN: 3BD # (MSPACE) 
$EXTMAX: 3BD # (MSPACE)
$LIMMIN: 2RD # (MSPACE) 
$LIMMAX: 2RD # (MSPACE) 
$ELEVATION: BD # (MSPACE) 
$UCSORG: 3BD # (MSPACE) 
$UCSXDIR: 3BD # (MSPACE) 
$UCSYDIR: 3BD # (MSPACE) 
$UCSNAME: H # (MSPACE) (hard pointer) 

ver: R2000+ 
$UCSORTHOREF: H # (hard pointer) 
$UCSORTHOVIEW: BS 
$UCSBASE: H # (hard pointer) 
$UCSORGTOP: 3BD 
$UCSORGBOTTOM: 3BD 
$UCSORGLEFT: 3BD 
$UCSORGRIGHT: 3BD 
$UCSORGFRONT: 3BD 
$UCSORGBACK: 3BD 
$DIMPOST: TV 
$DIMAPOST: TV 

ver: R13-R14
$DIMTOL: B 
$DIMLIM: B 
$DIMTIH: B 
$DIMTOH: B 
$DIMSE1: B 
$DIMSE2: B 
$DIMALT: B 
$DIMTOFL: B 
$DIMSAH: B 
$DIMTIX: B 
$DIMSOXD: B 
$DIMALTD: RC 
$DIMZIN: RC 
$DIMSD1: B 
$DIMSD2: B 
$DIMTOLJ: RC 
$DIMJUST: RC 
$DIMFIT: RC 
$DIMUPT: B 
$DIMTZIN: RC  
$DIMALTZ: RC 
$DIMALTTZ: RC 
$DIMTAD: RC 
$DIMUNIT: BS 
$DIMAUNIT: BS 
$DIMDEC: BS 
$DIMTDEC: BS 
$DIMALTU: BS 
$DIMALTTD: BS 
$DIMTXSTY: H # (hard pointer) 

ver: all 
$DIMSCALE: BD 
$DIMASZ: BD 
$DIMEXO: BD 
$DIMDLI: BD 
$DIMEXE: BD 
$DIMRND: BD 
$DIMDLE: BD 
$DIMTP: BD 
$DIMTM: BD 

ver: R2007+ 
$DIMFXL: BD 
$DIMJOGANG: BD 
$DIMTFILL: BS 
$DIMTFILLCLR: CMC 

ver: R2000+
$DIMTOL: B 
$DIMLIM: B 
$DIMTIH: B 
$DIMTOH: B 
$DIMSE1: B 
$DIMSE2: B 
$DIMTAD: BS 
$DIMZIN: BS 
$DIMAZIN: BS 

ver: R2007+ 
$DIMARCSYM: BS 

ver: all 
$DIMTXT: BD 
$DIMCEN: BD 
$DIMTSZ: BD 
$DIMALTF: BD 
$DIMLFAC: BD 
$DIMTVP: BD 
$DIMTFAC: BD 
$DIMGAP: BD 

ver: R13-R14 
$DIMPOST: T 
$DIMAPOST: T 
$DIMBLK: T 
$DIMBLK1: T 
$DIMBLK2: T 

ver: R2000+ 
$DIMALTRND: BD 
$DIMALT: B 
$DIMALTD: BS 
$DIMTOFL: B 
$DIMSAH: B 
$DIMTIX: B 
$DIMSOXD: B 

ver: all 
$DIMCLRD: CMC 
$DIMCLRE: CMC 
$DIMCLRT: CMC 

ver: R2000+
$DIMADEC: BS 
$DIMDEC: BS 
$DIMTDEC: BS 
$DIMALTU: BS 
$DIMALTTD: BS 
$DIMAUNIT: BS 
$DIMFRAC: BS 
$DIMLUNIT: BS 
$DIMDSEP: BS 
$DIMTMOVE: BS 
$DIMJUST: BS 
$DIMSD1: B 
$DIMSD2: B 
$DIMTOLJ: BS 
$DIMTZIN: BS 
$DIMALTZ: BS 
$DIMALTTZ: BS 
$DIMUPT: B 
$DIMATFIT: BS 

ver: R2007+
$DIMFXLON: B 

ver: R2010+ 
$DIMTXTDIRECTION: B 
$DIMALTMZF: BD 
$DIMALTMZS: T 
$DIMMZF: BD 
$DIMMZS: T 

ver: R2000+ 
$DIMTXSTY: H # (hard pointer) 
$DIMLDRBLK: H # (hard pointer) 
$DIMBLK: H # (hard pointer) 
$DIMBLK1: H # (hard pointer) 
$DIMBLK2: H # (hard pointer) 

ver: R2007+ 
$DIMLTYPE: H # (hard pointer) 
$DIMLTEX1: H # (hard pointer) 
$DIMLTEX2: H # (hard pointer) 

ver: R2000+ 
$DIMLWD: BS 
$DIMLWE: BS 

ver: all 
$BLOCK_CONTROL_OBJECT: H # (hard owner) Block Record Table
$LAYER_CONTROL_OBJECT: H # (hard owner) Layer Table
$STYLE_CONTROL_OBJECT: H # (hard owner) Style Table
$LINETYPE_CONTROL_OBJECT: H # (hard owner) Linetype Table
$VIEW_CONTROL_OBJECT: H # (hard owner) View table
$UCS_CONTROL_OBJECT: H # (hard owner)  UCS Table
$VPORT_CONTROL_OBJECT: H # (hard owner)  Viewport table
$APPID_CONTROL_OBJECT: H # (hard owner)  AppID Table
$DIMSTYLE_CONTROL_OBJECT: H # (hard owner)  Dimstyle Table

ver: R13-R2000 
$VIEWPORT_ENTITY_HEADER_CONTROL_OBJECT: H # (hard owner) 

ver: all 
$ACAD_GROUP_DICTIONARY: H # (hard pointer) 
$ACAD_MLINESTYLE_DICTIONARY: H # (hard pointer) 
$ROOT_DICTIONARY: H # (NAMED OBJECTS) (hard owner) 

ver: R2000+ 
$TSTACKALIGN: BS # default = 1 (not present in DXF) 
$TSTACKSIZE: BS #  default = 70 (not present in DXF) 
$HYPERLINKBASE: TV 
$STYLESHEET: TV 
$LAYOUTS_DICTIONARY: H # (hard pointer) 
$PLOTSETTINGS_DICTIONARY: H # (hard pointer) 
$PLOTSTYLES_DICTIONARY: H # (hard pointer) 

ver: R2004+ 
$MATERIALS_DICTIONARY: H # (hard pointer) 
$COLORS_DICTIONARY: H # (hard pointer) 

ver: R2007+ 
$VISUALSTYLE_DICTIONARY: H # (hard pointer) 

ver: R2013+
$UNKNOWN: H # (hard pointer) 

ver: R2000+
$R2000_PLUS_FLAGS: BL
#    CELWEIGHT Flags & 0x001F 
#    ENDCAPS Flags & 0x0060 
#    JOINSTYLE Flags & 0x0180 
#    LWDISPLAY !(Flags & 0x0200) 
#    XEDIT !(Flags & 0x0400) 
#    EXTNAMES Flags & 0x0800 
#    PSTYLEMODE Flags & 0x2000 
#    OLESTARTUP Flags & 0x4000 
$INSUNITS: BS 
$CEPSNTYPE: BS

skip_next_if: header['$CEPSNTYPE'] != 3
$CPSNID: H # (present only if CEPSNTYPE == 3) (hard pointer)

$FINGERPRINTGUID: TV 
$VERSIONGUID: TV 

ver: R2004+ 
$SORTENTS: RC 
$INDEXCTL: RC 
$HIDETEXT: RC 
$XCLIPFRAME: RC # before R2010 the value can be 0 or 1 only. 
$DIMASSOC: RC 
$HALOGAP: RC 
$OBSCUREDCOLOR: BS 
$INTERSECTIONCOLOR: BS 
$OBSCUREDLTYPE: RC 
$INTERSECTIONDISPLAY: RC 
$PROJECTNAME: TV 

ver: all 
$PAPER_SPACE_BLOCK_RECORD: H # (hard pointer) 
$MODEL_SPACE_BLOCK_RECORD: H #  (hard pointer) 
$BYLAYER_LTYPE: H # (hard pointer) 
$BYBLOCK_LTYPE: H # (hard pointer) 
$CONTINUOUS_LTYPE: H # (hard pointer) 

ver: R2007+ 
$CAMERADISPLAY: B 
$UNKNOWN: BL 
$UNKNOWN: BL 
$UNKNOWN: BD 
$STEPSPERSEC: BD 
$STEPSIZE: BD 
$3DDWFPREC: BD 
$LENSLENGTH: BD 
$CAMERAHEIGHT: BD 
$SOLIDHIST: RC 
$SHOWHIST: RC 
$PSOLWIDTH: BD 
$PSOLHEIGHT: BD 
$LOFTANG1: BD 
$LOFTANG2: BD 
$LOFTMAG1: BD 
$LOFTMAG2: BD 
$LOFTPARAM: BS 
$LOFTNORMALS: RC 
$LATITUDE: BD 
$LONGITUDE: BD 
$NORTHDIRECTION: BD 
$TIMEZONE: BL 
$LIGHTGLYPHDISPLAY: RC 
$TILEMODELIGHTSYNCH: RC 
$DWFFRAME: RC 
$DGNFRAME: RC 
$UNKNOWN: B 
$INTERFERECOLOR: CMC 
$INTERFEREOBJVS: H # (hard pointer) 
$INTERFEREVPVS: H # (hard pointer) 
$CSHADOW: RC 
$UNKNOWN: BD 

ver: R14+ 
$UNKNOWN: BS # short (type 5/6 only) these do not seem to be required, 
$UNKNOWN: BS # short (type 5/6 only) even for type 5. 
$UNKNOWN: BS # short (type 5/6 only) 
$UNKNOWN: BS # short (type 5/6 only) 

"""



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
# Created 2019-02-13

# first factory
from . import factory

# basic classes
from .xdict import ExtensionDict
from .xdata import XData, EmbeddedObjects
from .appdata import AppData, Reactors
from .dxfentity import DXFEntity
from .dxfgfx import DXFGraphic, SeqEnd
from .dxfobj import DXFObject
from .dxfns import DXFNamespace, SubclassProcessor

# register management structures
from .dxfclass import DXFClass
from .table import TableHead

# register table entries
from .ltype import Linetype
from .layer import Layer
from .textstyle import Textstyle
from .dimstyle import DimStyle
from .view import View
from .vport import VPort
from .ucs import UCSTable
from .appid import AppID
from .blockrecord import BlockRecord

# register DXF objects R2000
from .dxfobj import XRecord, Placeholder, VBAProject, SortEntsTable
from .dictionary import Dictionary, DictionaryVar, DictionaryWithDefault
from .layout import DXFLayout
from .idbuffer import IDBuffer
from .sun import Sun
from .material import Material, MaterialCollection

# register DXF objects R2007
from .visualstyle import VisualStyle

# register entities R12
from .line import Line
from .point import Point
from .circle import Circle
from .arc import Arc
from .shape import Shape
from .solid import Solid, Face3d, Trace
from .text import Text
from .subentity import LinkedEntities, entity_linker
from .insert import Insert
from .block import Block, EndBlk
from .polyline import Polyline, Polyface, Polymesh, MeshVertexCache
from .attrib import Attrib, AttDef
from .dimension import Dimension, ArcDimension
from .dimstyleoverride import DimStyleOverride
from .viewport import Viewport

# register graphical entities R2000
from .lwpolyline import LWPolyline
from .ellipse import Ellipse
from .xline import XLine, Ray
from .mtext import MText
from .spline import Spline
from .mesh import Mesh, MeshData
from .hatch import (
    Hatch, BoundaryPaths, PolylinePath, EdgePath, LineEdge,
    ArcEdge, EllipseEdge, SplineEdge, Pattern, PatternLine, Gradient,
)
from .image import Image, ImageDef, Wipeout
from .underlay import (
    Underlay, UnderlayDefinition, PdfUnderlay, DgnUnderlay,
    DwfUnderlay,
)
from .leader import Leader
from .tolerance import Tolerance
from .helix import Helix
from .acis import (
    Body, Solid3d, Region, Surface, ExtrudedSurface,
    LoftedSurface, RevolvedSurface, SweptSurface,
)
from .mline import MLine, MLineVertex, MLineStyle, MLineStyleCollection
from .mleader import MLeader, MLeaderStyle, MLeaderStyleCollection

# register graphical entities R2004

# register graphical entities R2007

from .light import Light

# register graphical entities R2010

from .geodata import GeoData

# register graphical entities R2013

# register graphical entities R2018



# Created: 28.12.2018
# Copyright (c) 2018-2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.math import UCS
from ezdxf.lldxf.const import DXFValueError
from ezdxf.entities.dimstyleoverride import DimStyleOverride

from .dim_linear import LinearDimension
from .dim_radius import RadiusDimension
from .dim_diameter import DiameterDimension

if TYPE_CHECKING:
    from ezdxf.eztypes import Dimension, BaseDimensionRenderer


class DimensionRenderer:
    def dispatch(self, override: 'DimStyleOverride', ucs: 'UCS') -> 'BaseDimensionRenderer':
        dimension = override.dimension
        dim_type = dimension.dimtype

        if dim_type in (0, 1):
            return self.linear(dimension, ucs, override)
        elif dim_type == 2:
            return self.angular(dimension, ucs, override)
        elif dim_type == 3:
            return self.diameter(dimension, ucs, override)
        elif dim_type == 4:
            return self.radius(dimension, ucs, override)
        elif dim_type == 5:
            return self.angular3p(dimension, ucs, override)
        elif dim_type == 6:
            return self.ordinate(dimension, ucs, override)
        else:
            raise DXFValueError(f'Unknown DIMENSION type: {dim_type}')

    def linear(self, dimension: 'Dimension', ucs: 'UCS', override: 'DimStyleOverride' = None):
        """ Call renderer for linear dimension lines: horizontal, vertical and rotated """
        return LinearDimension(dimension, ucs, override)

    def angular(self, dimension: 'Dimension', ucs: 'UCS', override: 'DimStyleOverride' = None):
        raise NotImplementedError()

    def diameter(self, dimension: 'Dimension', ucs: 'UCS', override: 'DimStyleOverride' = None):
        """ Call renderer for diameter dimension """
        return DiameterDimension(dimension, ucs, override)

    def radius(self, dimension: 'Dimension', ucs: 'UCS', override: 'DimStyleOverride' = None):
        """ Call renderer for radius dimension """
        return RadiusDimension(dimension, ucs, override)

    def angular3p(self, dimension: 'Dimension', ucs: 'UCS', override: 'DimStyleOverride' = None):
        raise NotImplementedError()

    def ordinate(self, dimension: 'Dimension', ucs: 'UCS', override: 'DimStyleOverride' = None):
        raise NotImplementedError()



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Optional
import copy
from ezdxf.lldxf import validator
from ezdxf.math import NULLVEC
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, RETURN_DEFAULT,
    group_code_mapping
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER, DXF2010
from ezdxf.lldxf import const
from ezdxf.tools import set_flag_state
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import acdb_entity, elevation_to_z_axis
from .text import Text, acdb_text, acdb_text_group_codes
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, Tags, DXFNamespace, DXFEntity

__all__ = ['AttDef', 'Attrib']

# DXF Reference for ATTRIB is a total mess and incorrect, the AcDbText subclass
# for the ATTRIB entity is the same as for the TEXT entity, but the valign field
# from the 2nd AcDbText subclass of the TEXT entity is stored in the
# AcDbAttribute subclass:
attrib_fields = {
    # Version number: 0 = 2010
    'version': DXFAttr(280, default=0, dxfversion=DXF2010),

    # Tag string (cannot contain spaces):
    'tag': DXFAttr(
        2, default='',
        validator=validator.is_valid_attrib_tag,
        fixer=validator.fix_attrib_tag,
    ),

    # 1 = Attribute is invisible (does not appear)
    # 2 = This is a constant attribute
    # 4 = Verification is required on input of this attribute
    # 8 = Attribute is preset (no prompt during insertion)
    'flags': DXFAttr(70, default=0),

    # Field length (optional) (not currently used)
    'field_length': DXFAttr(73, default=0, optional=True),

    # Vertical text justification type (optional); see group code 73 in TEXT
    'valign': DXFAttr(
        74, default=0, optional=True,
        validator=validator.is_in_integer_range(0, 4),
        fixer=RETURN_DEFAULT,
    ),

    # Lock position flag. Locks the position of the attribute within the block
    # reference, example of double use of group codes in one sub class
    'lock_position': DXFAttr(
        280, default=0, dxfversion=DXF2010, optional=True,
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),
}

# ATTDEF has an additional field: 'prompt'
# DXF attribute definitions are immutable, a shallow copy is sufficient:
attdef_fields = dict(attrib_fields)
attdef_fields['prompt'] = DXFAttr(
    3, default='',
    validator=validator.is_valid_one_line_text,
    fixer=validator.fix_one_line_text,
)

acdb_attdef = DefSubclass('AcDbAttributeDefinition', attdef_fields)
acdb_attdef_group_codes = group_code_mapping(acdb_attdef)
acdb_attrib = DefSubclass('AcDbAttribute', attrib_fields)
acdb_attrib_group_codes = group_code_mapping(acdb_attrib)

# For XRECORD the tag order is important and group codes appear multiple times,
# therefore this attribute definition needs a special treatment!
acdb_attdef_xrecord = DefSubclass('AcDbXrecord', [
    # Duplicate record cloning flag (determines how to merge duplicate entries):
    # 1 = Keep existing
    ('cloning', DXFAttr(280, default=1)),

    # MText flag:
    # 2 = multiline attribute
    # 4 = constant multiline attribute definition
    ('mtext_flag', DXFAttr(70, default=0)),

    # isReallyLocked flag:
    #     0 = unlocked
    #     1 = locked
    ('really_locked', DXFAttr(
        70, default=0,
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    )),

    # Number of secondary attributes or attribute definitions:
    ('secondary_attribs_count', DXFAttr(70, default=0)),
    # Hard-pointer id of secondary attribute(s) or attribute definition(s):
    ('secondary_attribs_handle', DXFAttr(70, default=0)),

    # Alignment point of attribute or attribute definition:
    ('align_point', DXFAttr(10, xtype=XType.point3d, default=NULLVEC)),
    ('current_annotation_scale', DXFAttr(40, default=0)),

    # attribute or attribute definition tag string
    ('tag', DXFAttr(
        2, default='',
        validator=validator.is_valid_attrib_tag,
        fixer=validator.fix_attrib_tag,
    )),
])


# A special MTEXT entity can follow the ATTDEF and ATTRIB entity, which starts
# as a usual DXF entity with (0, 'MTEXT'), so processing can't be done here,
# because for ezdxf is this a separated Entity.
#
# The attached MTEXT entity: owner is None and handle is None
# Linked as attribute `attached_mtext`.
# I don't have seen this combination of entities in real world examples and is
# ignored by ezdxf for now.


class BaseAttrib(Text):
    XRECORD_DEF = acdb_attdef_xrecord

    def __init__(self):
        """ Default constructor """
        super().__init__()
        self.xrecord: Optional['Tags'] = None
        self.attached_mtext: Optional['DXFEntity'] = None

    def _copy_data(self, entity: 'BaseAttrib') -> None:
        """ Copy entity data, xrecord data and attached MTEXT are not stored
        in the entity database.

        """
        entity.xrecord = copy.deepcopy(self.xrecord)
        if self.attached_mtext:
            entity.attached_mtext = self.attached_mtext.copy()
            # attached mtext entity is not stored in the entity database
            # no further action required

    def link_entity(self, entity: 'DXFEntity'):
        self.attached_mtext = entity

    def export_dxf(self, tagwriter: 'TagWriter'):
        """ Export attached MTEXT entity. """
        super().export_dxf(tagwriter)
        if self.attached_mtext:
            # todo: export MTEXT attached to ATTRIB
            # Attached MTEXT has no handle and owner and can not exported
            # by the usual export process:
            # self.attached_mtext.export_dxf(tagwriter)
            raise NotImplemented('Attached MTEXT export')

    @property
    def is_const(self) -> bool:
        """ This is a constant attribute. """
        return bool(self.dxf.flags & const.ATTRIB_CONST)

    @is_const.setter
    def is_const(self, state: bool) -> None:
        """ This is a constant attribute. """
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_CONST,
                                        state)

    @property
    def is_invisible(self) -> bool:
        """ Attribute is invisible (does not appear). """
        return bool(self.dxf.flags & const.ATTRIB_INVISIBLE)

    @is_invisible.setter
    def is_invisible(self, state: bool) -> None:
        """ Attribute is invisible (does not appear). """
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_INVISIBLE,
                                        state)

    @property
    def is_verify(self) -> bool:
        """ Verification is required on input of this attribute.
        (CAD application feature)

        """
        return bool(self.dxf.flags & const.ATTRIB_VERIFY)

    @is_verify.setter
    def is_verify(self, state: bool) -> None:
        """ Verification is required on input of this attribute.
        (CAD application feature)

        """
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_VERIFY,
                                        state)

    @property
    def is_preset(self) -> bool:
        """ No prompt during insertion. (CAD application feature) """
        return bool(self.dxf.flags & const.ATTRIB_IS_PRESET)

    @is_preset.setter
    def is_preset(self, state: bool) -> None:
        """ No prompt during insertion. (CAD application feature) """
        self.dxf.flags = set_flag_state(self.dxf.flags, const.ATTRIB_IS_PRESET,
                                        state)


@register_entity
class AttDef(BaseAttrib):
    """ DXF ATTDEF entity """
    DXFTYPE = 'ATTDEF'
    # Don't add acdb_attdef_xrecord here:
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_text, acdb_attdef)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super(Text, self).load_dxf_attribs(processor)
        # Do not call Text loader.
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_text_group_codes, 2, recover=True)
            processor.fast_load_dxfattribs(
                dxf, acdb_attdef_group_codes, 3, recover=True)
            self.xrecord = processor.find_subclass(self.XRECORD_DEF.name)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ('insert', 'align_point'))

        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        # Text() writes 2x AcDbText which is not suitable for AttDef()
        self.export_acdb_entity(tagwriter)
        self.export_acdb_text(tagwriter)
        self.export_acdb_attdef(tagwriter)
        if self.xrecord:
            tagwriter.write_tags(self.xrecord)

    def export_acdb_attdef(self, tagwriter: 'TagWriter') -> None:
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_attdef.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'version', 'prompt', 'tag', 'flags', 'field_length', 'valign',
            'lock_position',
        ])


@register_entity
class Attrib(BaseAttrib):
    """ DXF ATTRIB entity """

    DXFTYPE = 'ATTRIB'
    # Don't add acdb_attdef_xrecord here:
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_text, acdb_attrib)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super(Text, self).load_dxf_attribs(processor)
        # Do not call Text loader.
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_text_group_codes, 2, recover=True)
            processor.fast_load_dxfattribs(
                dxf, acdb_attrib_group_codes, 3, recover=True)
            self.xrecord = processor.find_subclass(self.XRECORD_DEF.name)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ('insert', 'align_point'))
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        # Text() writes 2x AcDbText which is not suitable for AttDef()
        self.export_acdb_entity(tagwriter)
        self.export_acdb_attrib_text(tagwriter)
        self.export_acdb_attrib(tagwriter)
        if self.xrecord:
            tagwriter.write_tags(self.xrecord)

    def export_acdb_attrib_text(self, tagwriter: 'TagWriter') -> None:
        # Despite the similarities to TEXT, it is different to
        # Text.export_acdb_text():
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_text.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'insert', 'height', 'text', 'thickness', 'rotation', 'oblique',
            'style', 'width', 'halign', 'align_point', 'text_generation_flag',
            'extrusion',
        ])

    def export_acdb_attrib(self, tagwriter: 'TagWriter') -> None:
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_attrib.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'version', 'tag', 'flags', 'field_length', 'valign',
            'lock_position',
        ])



# Copyright (c) 2010-2020, Manfred Moitzi
# License: MIT License
from typing import Dict, Iterable
from ezdxf.math import Vec3
from ezdxf.math.bspline import global_bspline_interpolation, BSpline


class EulerSpiral:
    """
    This class represents an euler spiral (clothoid) for `curvature` (Radius of curvature).

    This is a parametric curve, which always starts at the origin = ``(0, 0)``.

    Args:
        curvature: radius of curvature

    """

    def __init__(self, curvature: float = 1.0):
        self.curvature = curvature  # Radius of curvature
        self.curvature_powers = [curvature ** power for power in range(19)]
        self._cache = {}  # type: Dict[float, Vec3] # coordinates cache

    def radius(self, t: float) -> float:
        """
        Get radius of circle at distance `t`.

        """
        if t > 0.:
            return self.curvature_powers[2] / t
        else:
            return 0.  # radius = infinite

    def tangent(self, t: float) -> Vec3:
        """
        Get tangent at distance `t` as :class.`Vec3` object.

        """
        angle = t ** 2 / (2. * self.curvature_powers[2])
        return Vec3.from_angle(angle)

    def distance(self, radius: float) -> float:
        """
        Get distance L from origin for `radius`.

        """
        return self.curvature_powers[2] / float(radius)

    def point(self, t: float) -> Vec3:
        """
        Get point at distance `t` as :class.`Vec3`.

        """

        def term(length_power, curvature_power, const):
            return t ** length_power / (const * self.curvature_powers[curvature_power])

        if t not in self._cache:
            y = term(3, 2, 6.) - term(7, 6, 336.) + term(11, 10, 42240.) - \
                term(15, 14, 9676800.) + term(19, 18, 3530096640.)
            x = t - term(5, 4, 40.) + term(9, 8, 3456.) - term(13, 12, 599040.) + \
                term(17, 16, 175472640.)
            self._cache[t] = Vec3(x, y)
        return self._cache[t]

    def approximate(self, length: float, segments: int) -> Iterable[Vec3]:
        """
        Approximate curve of length with line segments.

        Generates segments+1 vertices as :class:`Vec3` objects.

        """
        delta_l = float(length) / float(segments)
        yield Vec3(0, 0)
        for index in range(1, segments + 1):
            yield self.point(delta_l * index)

    def circle_center(self, t: float) -> Vec3:
        """
        Get circle center at distance `t`.

        .. versionchanged:: 0.10

                renamed from `circle_midpoint`

        """
        p = self.point(t)
        r = self.radius(t)
        return p + self.tangent(t).normalize(r).orthogonal()

    def bspline(self, length: float, segments: int = 10, degree: int = 3, method: str = 'uniform') -> BSpline:
        """
        Approximate euler spiral as B-spline.

        Args:
            length: length of euler spiral
            segments: count of fit points for B-spline calculation
            degree: degree of BSpline
            method: calculation method for parameter vector t

        Returns:
            :class:`BSpline`

        """
        fit_points = list(self.approximate(length, segments=segments))
        spline = global_bspline_interpolation(fit_points, degree, method=method)
        knots = [v * length for v in spline.knot_values()]  # scale knot values to length
        spline.basis.knots = knots
        return spline

    # backward compatibility
    circle_midpoint = circle_center



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Sequence, Tuple, Union, List, Dict
import array
import copy
from itertools import chain
from contextlib import contextmanager
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, RETURN_DEFAULT, group_code_mapping
)
from ezdxf.lldxf.const import (
    SUBCLASS_MARKER, DXF2000, DXFValueError, DXFStructureError,
)
from ezdxf.lldxf.packedtags import VertexArray, TagArray, TagList
from ezdxf.tools import take2
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter, DXFNamespace, Vertex, Tags, Matrix44,
    )

__all__ = ['Mesh', 'MeshData']

acdb_mesh = DefSubclass('AcDbSubDMesh', {
    'version': DXFAttr(71, default=2),
    'blend_crease': DXFAttr(
        72, default=0,
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),

    # 0 is no smoothing
    'subdivision_levels': DXFAttr(
        91, default=0,
        validator=validator.is_greater_or_equal_zero,
        fixer=RETURN_DEFAULT,
    ),
    # 92: Vertex count of level 0
    # 10: Vertex position, multiple entries
    # 93: Size of face list of level 0
    # 90: Face list item, >=3 possible
    #     90: length of face list
    #     90: 1st vertex index
    #     90: 2nd vertex index ...
    # 94: Edge count of level 0
    #     90: Vertex index of 1st edge
    #     90: Vertex index of 2nd edge
    # 95: Edge crease count of level 0
    #     95 same as 94, or how is the 'edge create value' associated to edge index
    # 140: Edge crease value
    #
    # Overriding properties: how does this work?
    # 90: Count of sub-entity which property has been overridden
    # 91: Sub-entity marker
    # 92: Count of property was overridden
    # 90: Property type
    #     0 = Color
    #     1 = Material
    #     2 = Transparency
    #     3 = Material mapper
})
acdb_mesh_group_codes = group_code_mapping(acdb_mesh)


class EdgeArray(TagArray):
    DTYPE = 'L'

    def __len__(self) -> int:
        return len(self.values) // 2

    def __iter__(self) -> Iterable[Tuple[int, int]]:
        for edge in take2(self.values):
            yield edge

    def set_data(self, edges: Iterable[Tuple[int, int]]) -> None:
        self.values = array.array(self.DTYPE, chain.from_iterable(edges))

    def export_dxf(self, tagwriter: 'TagWriter'):
        # count = count of edges not tags!
        tagwriter.write_tag2(94, len(self.values) // 2)
        for index in self.values:
            tagwriter.write_tag2(90, index)


class FaceList(TagList):
    def __len__(self) -> int:
        return len(self.values)

    def __iter__(self) -> Iterable[array.array]:
        return iter(self.values)

    def export_dxf(self, tagwriter: 'TagWriter'):
        # count = count of tags not faces!
        tagwriter.write_tag2(93, self.tag_count())
        for face in self.values:
            tagwriter.write_tag2(90, len(face))
            for index in face:
                tagwriter.write_tag2(90, index)

    def tag_count(self) -> int:
        return len(self.values) + sum(len(f) for f in self.values)

    def set_data(self, faces: Iterable[Sequence[int]]) -> None:
        _faces = []
        for face in faces:
            _faces.append(face_to_array(face))
        self.values = _faces


def face_to_array(face: Sequence[int]) -> array.array:
    max_index = max(face)
    if max_index < 256:
        dtype = 'B'
    elif max_index < 65536:
        dtype = 'I'
    else:
        dtype = 'L'
    return array.array(dtype, face)


def create_vertex_array(tags: 'Tags', start_index: int) -> 'VertexArray':
    vertex_tags = tags.collect_consecutive_tags(codes=(10,), start=start_index)
    return VertexArray(data=chain.from_iterable(t.value for t in vertex_tags))


def create_face_list(tags: 'Tags', start_index: int) -> 'FaceList':
    faces = FaceList()
    faces_list = faces.values
    face = []
    counter = 0
    for tag in tags.collect_consecutive_tags(codes=(90,), start=start_index):
        if not counter:
            # leading counter tag
            counter = tag.value
            if face:
                # group code 90 = 32 bit integer
                faces_list.append(face_to_array(face))
                face = []
        else:
            # followed by count face tags
            counter -= 1
            face.append(tag.value)

    # add last face
    if face:
        # group code 90 = 32 bit integer
        faces_list.append(face_to_array(face))

    return faces


def create_edge_array(tags: 'Tags', start_index: int) -> 'EdgeArray':
    return EdgeArray(data=collect_values(
        tags, start_index, code=90))  # int values


def collect_values(tags: 'Tags',
                   start_index: int,
                   code: int) -> Iterable[Union[float, int]]:
    values = tags.collect_consecutive_tags(codes=(code,), start=start_index)
    return (t.value for t in values)


def create_crease_array(tags: 'Tags', start_index: int) -> 'array.array':
    return array.array('f', collect_values(
        tags, start_index, code=140))  # float values


COUNT_ERROR_MSG = "'MESH (#{}) without {} count.'"


@register_entity
class Mesh(DXFGraphic):
    """ DXF MESH entity """
    DXFTYPE = 'MESH'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_mesh)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def __init__(self):
        super().__init__()
        self._vertices = VertexArray()  # vertices stored as array.array('d')
        self._faces = FaceList()  # face lists data
        self._edges = EdgeArray()  # edge indices stored as array.array('L')
        self._creases = array.array('f')  # creases stored as array.array('f')

    def _copy_data(self, entity: 'Mesh') -> None:
        """ Copy data: vertices, faces, edges, creases. """
        entity._vertices = copy.deepcopy(self._vertices)
        entity._faces = copy.deepcopy(self._faces)
        entity._edges = copy.deepcopy(self._edges)
        entity._creases = copy.deepcopy(self._creases)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.subclass_by_index(2)
            # Load mesh data and remove their tags from subclass
            self.load_mesh_data(tags, dxf.handle)
            # Load remaining data into name space
            processor.fast_load_dxfattribs(
                dxf, acdb_mesh_group_codes, 2, recover=True)
        return dxf

    def load_mesh_data(self, mesh_tags: 'Tags', handle: str) -> None:
        def process_vertices():
            try:
                vertex_count_index = mesh_tags.tag_index(92)
            except DXFValueError:
                raise DXFStructureError(
                    COUNT_ERROR_MSG.format(handle, 'vertex'))
            vertices = create_vertex_array(mesh_tags, vertex_count_index + 1)
            # Remove vertex count tag and all vertex tags
            end_index = vertex_count_index + 1 + len(vertices)
            del mesh_tags[vertex_count_index:end_index]
            return vertices

        def process_faces():
            try:
                face_count_index = mesh_tags.tag_index(93)
            except DXFValueError:
                raise DXFStructureError(COUNT_ERROR_MSG.format(handle, 'face'))
            else:
                # Remove face count tag and all face tags
                faces = create_face_list(mesh_tags, face_count_index + 1)
                end_index = face_count_index + 1 + faces.tag_count()
                del mesh_tags[face_count_index:end_index]
                return faces

        def process_edges():
            try:
                edge_count_index = mesh_tags.tag_index(94)
            except DXFValueError:
                raise DXFStructureError(COUNT_ERROR_MSG.format(handle, 'edge'))
            else:
                edges = create_edge_array(mesh_tags, edge_count_index + 1)
                # Remove edge count tag and all edge tags
                end_index = edge_count_index + 1 + len(edges.values)
                del mesh_tags[edge_count_index:end_index]
                return edges

        def process_creases():
            try:
                crease_count_index = mesh_tags.tag_index(95)
            except DXFValueError:
                raise DXFStructureError(
                    COUNT_ERROR_MSG.format(handle, 'crease'))
            else:
                creases = create_crease_array(mesh_tags, crease_count_index + 1)
                # Remove crease count tag and all crease tags
                end_index = crease_count_index + 1 + len(creases)
                del mesh_tags[crease_count_index:end_index]
                return creases

        self._vertices = process_vertices()
        self._faces = process_faces()
        self._edges = process_edges()
        self._creases = process_creases()

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_mesh.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'version', 'blend_crease', 'subdivision_levels'
        ])
        self.export_mesh_data(tagwriter)
        self.export_override_data(tagwriter)

    def export_mesh_data(self, tagwriter: 'TagWriter'):
        tagwriter.write_tag2(92, len(self.vertices))
        self._vertices.export_dxf(tagwriter, code=10)
        self._faces.export_dxf(tagwriter)
        self._edges.export_dxf(tagwriter)

        tagwriter.write_tag2(95, len(self.creases))
        for crease_value in self.creases:
            tagwriter.write_tag2(140, crease_value)

    def export_override_data(self, tagwriter: 'TagWriter'):
        tagwriter.write_tag2(90, 0)

    @property
    def creases(self) -> 'array.array':
        """ Creases as :class:`array.array`. (read/write)"""
        return self._creases

    @creases.setter
    def creases(self, values: Iterable[float]) -> None:
        self._creases = array.array('f', values)

    @property
    def vertices(self):
        """ Vertices as list like :class:`~ezdxf.lldxf.packedtags.VertexArray`.
        (read/write)
        """
        return self._vertices

    @vertices.setter
    def vertices(self, points: Iterable['Vertex']) -> None:
        self._vertices = VertexArray(chain.from_iterable(points))

    @property
    def edges(self):
        """ Edges as list like :class:`~ezdxf.lldxf.packedtags.TagArray`.
        (read/write)
        """
        return self._edges

    @edges.setter
    def edges(self, edges: Iterable[Tuple[int, int]]) -> None:
        self._edges.set_data(edges)

    @property
    def faces(self):
        """ Faces as list like :class:`~ezdxf.lldxf.packedtags.TagList`.
        (read/write)
        """
        return self._faces

    @faces.setter
    def faces(self, faces: Iterable[Sequence[int]]) -> None:
        self._faces.set_data(faces)

    def get_data(self) -> 'MeshData':
        return MeshData(self)

    def set_data(self, data: 'MeshData') -> None:
        self.vertices = data.vertices
        self._faces.set_data(data.faces)
        self._edges.set_data(data.edges)
        self.creases = data.edge_crease_values

    @contextmanager
    def edit_data(self) -> 'MeshData':
        """ Context manager various mesh data, returns :class:`MeshData`.

        Despite that vertices, edge and faces since `ezdxf` v0.8.9 are
        accessible as packed data types, the usage of :class:`MeshData`
        by context manager :meth:`edit_data` is still recommended.

        """
        data = self.get_data()
        yield data
        self.set_data(data)

    def transform(self, m: 'Matrix44') -> 'Mesh':
        """ Transform MESH entity by transformation matrix `m` inplace.

        .. versionadded:: 0.13

        """
        self._vertices.transform(m)
        return self


class MeshData:
    def __init__(self, mesh):
        self.vertices: List[Tuple[float, float, float]] = list(mesh.vertices)
        self.faces: List[array.array] = list(mesh.faces)
        self.edges: List[Tuple[int, int]] = list(mesh.edges)
        self.edge_crease_values: array.array = mesh.creases

    def add_face(self, vertices: Iterable[Sequence[float]]) -> Sequence[int]:
        """ Add a face by coordinates, vertices is a list of ``(x, y, z)``
        tuples.
        """
        return self.add_entity(vertices, self.faces)

    def add_edge(self, vertices: Sequence[Sequence[float]]) -> Sequence[int]:
        """ Add an edge by coordinates, vertices is a list of two ``(x, y, z)``
        tuples.
        """
        if len(vertices) != 2:
            raise DXFValueError(
                "Parameter vertices has to be a list/tuple of 2 vertices "
                "[(x1, y1, z1), (x2, y2, z2)].")
        return self.add_entity(vertices, self.edges)

    def add_entity(self, vertices: Iterable[Sequence[float]],
                   entity_list: List) -> Sequence[int]:
        indices = [self.add_vertex(vertex) for vertex in vertices]
        entity_list.append(indices)
        return indices

    def add_vertex(self, vertex: Sequence[float]) -> int:
        if len(vertex) != 3:
            raise DXFValueError(
                'Parameter vertex has to be a 3-tuple (x, y, z).')
        index = len(self.vertices)
        self.vertices.append(vertex)
        return index

    def optimize(self, precision: int = 6):
        """
        Try to reduce vertex count by merging near vertices. `precision`
        defines the decimal places for coordinate be equal to merge two vertices.

        """

        def remove_doublette_vertices() -> Dict[int, int]:
            def prepare_vertices() -> Iterable[Tuple[float, float, float]]:
                for index, vertex in enumerate(self.vertices):
                    x, y, z = vertex
                    yield (round(x, precision),
                           round(y, precision),
                           round(z, precision),
                           index)

            sorted_vertex_list = list(sorted(prepare_vertices()))
            original_vertices = self.vertices
            self.vertices = []
            index_map: Dict[int, int] = {}
            cmp_vertex = None
            index = 0
            while len(sorted_vertex_list):
                vertex_entry = sorted_vertex_list.pop()
                original_index = vertex_entry[3]
                vertex = original_vertices[original_index]
                if vertex != cmp_vertex:
                    # this is not a doublette
                    index = len(self.vertices)
                    self.vertices.append(vertex)
                    index_map[original_index] = index
                    cmp_vertex = vertex
                else:  # it is a doublette
                    index_map[original_index] = index
            return index_map

        def remap_faces() -> None:
            self.faces = remap_indices(self.faces)

        def remap_edges() -> None:
            self.edges = remap_indices(self.edges)

        def remap_indices(entity_list: Sequence[Sequence[int]]) -> List[Tuple]:
            mapped_indices = []  # type: List[Tuple]
            for entity in entity_list:
                index_list = [index_map[index] for index in entity]
                mapped_indices.append(tuple(index_list))
            return mapped_indices

        index_map = remove_doublette_vertices()
        remap_faces()
        remap_edges()



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000
from ezdxf.math import NULLVEC, Z_AXIS, X_AXIS
from ezdxf.math.transformtools import transform_extrusion
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace, Matrix44

__all__ = ['Tolerance']

acdb_tolerance = DefSubclass('AcDbFcf', {
    'dimstyle': DXFAttr(
        3, default='Standard',
        validator=validator.is_valid_table_name,
    ),
    # Insertion point (in WCS):
    'insert': DXFAttr(10, xtype=XType.point3d, default=NULLVEC),

    # String representing the visual representation of the tolerance:
    'content': DXFAttr(1, default=""),
    'extrusion': DXFAttr(
        210, xtype=XType.point3d, default=Z_AXIS, optional=True,
        validator=validator.is_not_null_vector,
        fixer=RETURN_DEFAULT,
    ),
    # X-axis direction vector (in WCS):
    'x_axis_vector': DXFAttr(
        11, xtype=XType.point3d, default=X_AXIS,
        validator=validator.is_not_null_vector,
        fixer=RETURN_DEFAULT,
    ),

})
acdb_tolerance_group_codes = group_code_mapping(acdb_tolerance)


@register_entity
class Tolerance(DXFGraphic):
    """ DXF TOLERANCE entity """
    DXFTYPE = 'TOLERANCE'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_tolerance)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_tolerance_group_codes, subclass=2, recover=True)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_tolerance.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'dimstyle', 'insert', 'content', 'extrusion', 'x_axis_vector'
        ])

    def transform(self, m: 'Matrix44') -> 'Tolerance':
        """ Transform TOLERANCE entity by transformation matrix `m` inplace.

        .. versionadded:: 0.13

        """
        self.dxf.insert = m.transform(self.dxf.insert)
        self.dxf.x_axis_vector = m.transform_direction(self.dxf.x_axis_vector)
        self.dxf.extrusion, _ = transform_extrusion(self.dxf.extrusion, m)
        return self



# Copyright (c) 2011-2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterator, Iterable, Union, cast
from collections import Counter, OrderedDict

from ezdxf.lldxf.const import DXFStructureError, DXF2004, DXF2000, DXFKeyError
from ezdxf.entities.dxfclass import DXFClass
from ezdxf.entities.dxfentity import DXFEntity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, Drawing, DXFEntity, DXFTagStorage

# name: cpp_class_name (2), app_name (3), flags(90), was_a_proxy (280),
# is_an_entity (281)
# Multiple entries for 'name' are possible and supported, ClassSection stores
# entries with key: (name, cpp_class_name).

CLASS_DEFINITIONS = {
    'ACDBDICTIONARYWDFLT': [
        'AcDbDictionaryWithDefault', 'ObjectDBX Classes', 0, 0, 0],
    'SUN': [
        'AcDbSun', 'SCENEOE', 1153, 0, 0],
    'DICTIONARYVAR': [
        'AcDbDictionaryVar', 'ObjectDBX Classes', 0, 0, 0],
    'TABLESTYLE': [
        'AcDbTableStyle', 'ObjectDBX Classes', 4095, 0, 0],
    'MATERIAL': [
        'AcDbMaterial', 'ObjectDBX Classes', 1153, 0, 0],
    'VISUALSTYLE': [
        'AcDbVisualStyle', 'ObjectDBX Classes', 4095, 0, 0],
    'SCALE': [
        'AcDbScale', 'ObjectDBX Classes', 1153, 0, 0],
    'MLEADERSTYLE': [
        'AcDbMLeaderStyle', 'ACDB_MLEADERSTYLE_CLASS', 4095, 0, 0],
    'MLEADER': [
        'AcDbMLeader', 'ACDB_MLEADER_CLASS', 1025, 0, 1],
    'CELLSTYLEMAP': [
        'AcDbCellStyleMap', 'ObjectDBX Classes', 1152, 0, 0],
    'EXACXREFPANELOBJECT': [
        'ExAcXREFPanelObject', 'EXAC_ESW', 1025, 0, 0],
    'NPOCOLLECTION': [
        'AcDbImpNonPersistentObjectsCollection',
        'ObjectDBX Classes', 1153, 0, 0],
    'LAYER_INDEX': [
        'AcDbLayerIndex', 'ObjectDBX Classes', 0, 0, 0],
    'SPATIAL_INDEX': [
        'AcDbSpatialIndex', 'ObjectDBX Classes', 0, 0, 0],
    'IDBUFFER': [
        'AcDbIdBuffer', 'ObjectDBX Classes', 0, 0, 0],
    'DIMASSOC': [
        'AcDbDimAssoc',
        '"AcDbDimAssoc|Product Desc:     AcDim ARX App For Dimension'
        '|Company:          Autodesk, Inc.'
        '|WEB Address:      www.autodesk.com"',
        0, 0, 0],
    'ACDBSECTIONVIEWSTYLE': [
        'AcDbSectionViewStyle', 'ObjectDBX Classes', 1025, 0, 0],
    'ACDBDETAILVIEWSTYLE': [
        'AcDbDetailViewStyle', 'ObjectDBX Classes', 1025, 0, 0],
    'IMAGEDEF': [
        'AcDbRasterImageDef', 'ISM', 0, 0, 0],
    'RASTERVARIABLES': [
        'AcDbRasterVariables', 'ISM', 0, 0, 0],
    'IMAGEDEF_REACTOR': [
        'AcDbRasterImageDefReactor', 'ISM', 1, 0, 0],
    'IMAGE': [
        'AcDbRasterImage', 'ISM', 2175, 0, 1],
    'PDFDEFINITION': [
        'AcDbPdfDefinition', 'ObjectDBX Classes', 1153, 0, 0],
    'PDFUNDERLAY': [
        'AcDbPdfReference', 'ObjectDBX Classes', 4095, 0, 1],
    'DWFDEFINITION': [
        'AcDbDwfDefinition', 'ObjectDBX Classes', 1153, 0, 0],
    'DWFUNDERLAY': [
        'AcDbDwfReference', 'ObjectDBX Classes', 1153, 0, 1],
    'DGNDEFINITION': [
        'AcDbDgnDefinition', 'ObjectDBX Classes', 1153, 0, 0],
    'DGNUNDERLAY': [
        'AcDbDgnReference', 'ObjectDBX Classes', 1153, 0, 1],
    'MENTALRAYRENDERSETTINGS': [
        'AcDbMentalRayRenderSettings', 'SCENEOE', 1024, 0, 0],
    'ACDBPLACEHOLDER': [
        'AcDbPlaceHolder', 'ObjectDBX Classes', 0, 0, 0],
    'LAYOUT': [
        'AcDbLayout', 'ObjectDBX Classes', 0, 0, 0],
    'SURFACE': [
        'AcDbSurface', 'ObjectDBX Classes', 4095, 0, 1],
    'EXTRUDEDSURFACE': [
        'AcDbExtrudedSurface', 'ObjectDBX Classes', 4095, 0, 1],
    'LOFTEDSURFACE': [
        'AcDbLoftedSurface', 'ObjectDBX Classes', 0, 0, 1],
    'REVOLVEDSURFACE': [
        'AcDbRevolvedSurface', 'ObjectDBX Classes', 0, 0, 1],
    'SWEPTSURFACE': [
        'AcDbSweptSurface', 'ObjectDBX Classes', 0, 0, 1],
    'PLANESURFACE': [
        'AcDbPlaneSurface', 'ObjectDBX Classes', 4095, 0, 1],
    'NURBSSURFACE': [
        'AcDbNurbSurface', 'ObjectDBX Classes', 4095, 0, 1],
    'ACDBASSOCEXTRUDEDSURFACEACTIONBODY': [
        'AcDbAssocExtrudedSurfaceActionBody', 'ObjectDBX Classes', 1025, 0, 0],
    'ACDBASSOCLOFTEDSURFACEACTIONBODY': [
        'AcDbAssocLoftedSurfaceActionBody', 'ObjectDBX Classes', 1025, 0, 0],
    'ACDBASSOCREVOLVEDSURFACEACTIONBODY': [
        'AcDbAssocRevolvedSurfaceActionBody', 'ObjectDBX Classes', 1025, 0, 0],
    'ACDBASSOCSWEPTSURFACEACTIONBODY': [
        'AcDbAssocSweptSurfaceActionBody', 'ObjectDBX Classes', 1025, 0, 0],
    'HELIX': [
        'AcDbHelix', 'ObjectDBX Classes', 4095, 0, 1],
    'WIPEOUT': [
        'AcDbWipeout', 'WipeOut|AutoCAD Express Tool|expresstools@autodesk.com',
        127, 0, 1],
    'WIPEOUTVARIABLES': [
        'AcDbWipeoutVariables',
        'WipeOut|AutoCAD Express Tool|expresstools@autodesk.com',
        0, 0, 0],
    'FIELDLIST': [
        'AcDbFieldList', 'ObjectDBX Classes', 1152, 0, 0],
    'GEODATA': [
        'AcDbGeoData', 'ObjectDBX Classes', 4095, 0, 0],
    'SORTENTSTABLE': [
        'AcDbSortentsTable', 'ObjectDBX Classes', 0, 0, 0],
    'ACAD_TABLE': [
        'AcDbTable', 'ObjectDBX Classes', 1025, 0, 1],
    'ARC_DIMENSION': [
        'AcDbArcDimension', 'ObjectDBX Classes', 1025, 0, 1],
    'LARGE_RADIAL_DIMENSION': [
        'AcDbRadialDimensionLarge', 'ObjectDBX Classes', 1025, 0, 1],
}

REQ_R2000 = [
    'ACDBDICTIONARYWDFLT', 'SUN', 'VISUALSTYLE', 'MATERIAL', 'SCALE',
    'TABLESTYLE', 'MLEADERSTYLE', 'DICTIONARYVAR', 'CELLSTYLEMAP',
    'MENTALRAYRENDERSETTINGS', 'ACDBDETAILVIEWSTYLE', 'ACDBSECTIONVIEWSTYLE',
    'RASTERVARIABLES', 'ACDBPLACEHOLDER', 'LAYOUT'
]

REQ_R2004 = [
    'ACDBDICTIONARYWDFLT', 'SUN', 'VISUALSTYLE', 'MATERIAL', 'SCALE',
    'TABLESTYLE', 'MLEADERSTYLE', 'DICTIONARYVAR', 'CELLSTYLEMAP',
    'MENTALRAYRENDERSETTINGS', 'ACDBDETAILVIEWSTYLE', 'ACDBSECTIONVIEWSTYLE',
    'RASTERVARIABLES'
]

REQUIRED_CLASSES = {
    DXF2000: REQ_R2000,
    DXF2004: REQ_R2004,
}


class ClassesSection:
    def __init__(self, doc: 'Drawing' = None,
                 entities: Iterable[DXFEntity] = None):
        # Multiple entries for 'name' possible -> key is (name, cpp_class_name)
        # DXFClasses are not stored in the entities database, because CLASS has
        # no handle.
        self.classes = OrderedDict()
        self.doc = doc
        if entities is not None:
            self.load(iter(entities))

    def __iter__(self) -> Iterable[DXFClass]:
        return (cls for cls in self.classes.values())

    def load(self, entities: Iterator[DXFEntity]) -> None:
        section_head: 'DXFTagStorage' = cast('DXFTagStorage', next(entities))

        if section_head.dxftype() != 'SECTION' or \
                section_head.base_class[1] != (2, 'CLASSES'):
            raise DXFStructureError(
                "Critical structure error in CLASSES section.")

        for cls_entity in entities:
            self.register(cast(DXFClass, cls_entity))

    def register(self,
                 classes: Union[DXFClass, Iterable[DXFClass]] = None) -> None:
        if classes is None:
            return

        if isinstance(classes, DXFClass):
            classes = (classes,)

        for dxfclass in classes:
            key = dxfclass.key
            if key not in self.classes:
                self.classes[key] = dxfclass

    def add_class(self, name: str):
        """ Register a known class by `name`. """
        if name not in CLASS_DEFINITIONS:
            return
        cls_data = CLASS_DEFINITIONS[name]
        cls = DXFClass.new(doc=self.doc)
        cpp, app, flags, proxy, entity = cls_data
        cls.update_dxf_attribs({
            'name': name,
            'cpp_class_name': cpp,
            'app_name': app,
            'flags': flags,
            'was_a_proxy': proxy,
            'is_an_entity': entity,
        })
        self.register(cls)

    def get(self, name: str) -> DXFClass:
        """ Returns the first class matching `name`.

        Storage key is the ``(name, cpp_class_name)`` tuple, because there are
        some classes with the same :attr:`name` but different
        :attr:`cpp_class_names`.

        """
        for cls in self.classes.values():
            if cls.dxf.name == name:
                return cls
        raise DXFKeyError(name)

    def add_required_classes(self, dxfversion: str) -> None:
        """ Add all required CLASS definitions for `dxfversion`. """
        names = REQUIRED_CLASSES.get(dxfversion, REQ_R2004)
        for name in names:
            self.add_class(name)

        if self.doc is None:  # testing environment SUT
            return

        dxf_types_in_use = self.doc.entitydb.dxf_types_in_use()
        if 'IMAGE' in dxf_types_in_use:
            self.add_class('IMAGE')
            self.add_class('IMAGEDEF')
            self.add_class('IMAGEDEF_REACTOR')
        if 'PDFUNDERLAY' in dxf_types_in_use:
            self.add_class('PDFDEFINITION')
            self.add_class('PDFUNDERLAY')
        if 'DWFUNDERLAY' in dxf_types_in_use:
            self.add_class('DWFDEFINITION')
            self.add_class('DWFUNDERLAY')
        if 'DGNUNDERLAY' in dxf_types_in_use:
            self.add_class('DGNDEFINITION')
            self.add_class('DGNUNDERLAY')
        if 'EXTRUDEDSURFACE' in dxf_types_in_use:
            self.add_class('EXTRUDEDSURFACE')
            self.add_class('ACDBASSOCEXTRUDEDSURFACEACTIONBODY')
        if 'LOFTEDSURFACE' in dxf_types_in_use:
            self.add_class('LOFTEDSURFACE')
            self.add_class('ACDBASSOCLOFTEDSURFACEACTIONBODY')
        if 'REVOLVEDSURFACE' in dxf_types_in_use:
            self.add_class('REVOLVEDSURFACE')
            self.add_class('ACDBASSOCREVOLVEDSURFACEACTIONBODY')
        if 'SWEPTSURFACE' in dxf_types_in_use:
            self.add_class('SWEPTSURFACE')
            self.add_class('ACDBASSOCSWEPTSURFACEACTIONBODY')

        for dxftype in dxf_types_in_use:
            self.add_class(dxftype)

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        """ Export DXF tags. (internal API) """
        tagwriter.write_str("  0\nSECTION\n  2\nCLASSES\n")
        for dxfclass in self.classes.values():
            dxfclass.export_dxf(tagwriter)
        tagwriter.write_str("  0\nENDSEC\n")

    def update_instance_counters(self) -> None:
        """ Update CLASS instance counter for all registered classes, requires
        DXF R2004+.
        """
        if self.doc.dxfversion < DXF2004:
            return  # instance counter not supported
        counter = Counter()
        # count all entities in the entity database
        for entity in self.doc.entitydb.values():
            counter[entity.dxftype()] += 1

        for dxfclass in self.classes.values():
            dxfclass.dxf.instance_count = counter[dxfclass.dxf.name]



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Tuple
import math
from ezdxf.math import Vec3, X_AXIS, Y_AXIS, Vec2, Matrix44, sign, OCS

if TYPE_CHECKING:
    from ezdxf.eztypes import DXFGraphic, Vertex


class TransformError(Exception):
    pass


class NonUniformScalingError(TransformError):
    pass


class InsertTransformationError(TransformError):
    pass


def transform_thickness_and_extrusion_without_ocs(entity: 'DXFGraphic', m: Matrix44) -> None:
    if entity.dxf.hasattr('thickness'):
        thickness = entity.dxf.thickness
        reflection = sign(thickness)
        thickness = m.transform_direction(entity.dxf.extrusion * thickness)
        entity.dxf.thickness = thickness.magnitude * reflection
        entity.dxf.extrusion = thickness.normalize()
    elif entity.dxf.hasattr('extrusion'):  # without thickness?
        extrusion = m.transform_direction(entity.dxf.extrusion)
        entity.dxf.extrusion = extrusion.normalize()


def transform_extrusion(extrusion: 'Vertex', m: Matrix44) -> Tuple[Vec3, bool]:
    """
    Transforms the old `extrusion` vector into a new extrusion vector. Returns the new extrusion vector and a
    boolean value: ``True`` if the new OCS established by the new extrusion vector has a uniform scaled xy-plane,
    else ``False``.

    The new extrusion vector is perpendicular to plane defined by the transformed x- and y-axis.

    Args:
        extrusion: extrusion vector of the old OCS
        m: transformation matrix

    Returns:

    """
    ocs = OCS(extrusion)
    ocs_x_axis_in_wcs = ocs.to_wcs(X_AXIS)
    ocs_y_axis_in_wcs = ocs.to_wcs(Y_AXIS)
    x_axis, y_axis = m.transform_directions((ocs_x_axis_in_wcs, ocs_y_axis_in_wcs))

    # Not sure if this is the correct test for a uniform scaled xy-plane
    is_uniform = math.isclose(x_axis.magnitude_square, y_axis.magnitude_square, abs_tol=1e-9)
    new_extrusion = x_axis.cross(y_axis).normalize()
    return new_extrusion, is_uniform


class OCSTransform:
    def __init__(self, extrusion: Vec3 = None, m: Matrix44 = None):
        self.m = m
        if extrusion is None:
            self.old_ocs = None
            self.scale_uniform = False
            self.new_ocs = None
        else:
            self.old_ocs = OCS(extrusion)
            new_extrusion, self.scale_uniform = transform_extrusion(extrusion, m)
            self.new_ocs = OCS(new_extrusion)

    @property
    def old_extrusion(self) -> Vec3:
        return self.old_ocs.uz

    @property
    def new_extrusion(self) -> Vec3:
        return self.new_ocs.uz

    @classmethod
    def from_ocs(cls, old: OCS, new: OCS, m: Matrix44) -> 'OCSTransform':
        ocs = cls()
        ocs.m = m
        ocs.old_ocs = old
        ocs.new_ocs = new
        return ocs

    def transform_length(self, length: 'Vertex', reflection=1.0) -> float:
        """ Returns magnitude of `length` direction vector transformed from
        old OCS into new OCS including `reflection` correction applied.
        """
        return self.m.transform_direction(self.old_ocs.to_wcs(length)).magnitude * sign(reflection)

    transform_scale_factor = transform_length

    def transform_vertex(self, vertex: 'Vertex') -> Vec3:
        """ Returns vertex transformed from old OCS into new OCS. """
        return self.new_ocs.from_wcs(self.m.transform(self.old_ocs.to_wcs(vertex)))

    def transform_2d_vertex(self, vertex: 'Vertex', elevation: float) -> Vec2:
        """ Returns 2D vertex transformed from old OCS into new OCS. """
        v = Vec3(vertex).replace(z=elevation)
        return self.new_ocs.from_wcs(self.m.transform(self.old_ocs.to_wcs(v))).vec2

    def transform_direction(self, direction: 'Vertex') -> Vec3:
        """ Returns direction transformed from old OCS into new OCS. """
        return self.new_ocs.from_wcs(self.m.transform_direction(self.old_ocs.to_wcs(direction)))

    def transform_angle(self, angle: float) -> float:
        """ Returns angle (in radians) from old OCS transformed into new OCS.
        """
        return self.transform_direction(Vec3.from_angle(angle)).angle % math.tau

    def transform_deg_angle(self, angle: float) -> float:
        """ Returns angle (in degrees) from old OCS transformed into new OCS.
        """
        return math.degrees(self.transform_angle(math.radians(angle))) % 360.0



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List, Iterable
from ezdxf.lldxf import validator
from ezdxf.lldxf import const
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.types import DXFTag, DXFVertex
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.const import (
    DXF12, SUBCLASS_MARKER, DXFStructureError, DXFValueError,
    DXFTableEntryError,
)
from ezdxf.math import Vec3, NULLVEC, X_AXIS, Y_AXIS
from ezdxf.tools import set_flag_state
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace, Drawing

__all__ = ['Viewport']

acdb_viewport = DefSubclass('AcDbViewport', {
    # DXF reference: Center point (in WCS)
    # Correction to the DXF reference:
    # This point represents the center of the viewport in paper space units
    # (DCS), but is stored as 3D point inclusive z-axis!
    'center': DXFAttr(10, xtype=XType.point3d, default=NULLVEC),

    # Width in paper space units:
    'width': DXFAttr(40, default=1),

    # Height in paper space units:
    'height': DXFAttr(41, default=1),

    # Viewport status field:
    # -1 = On, but is fully off screen, or is one of the viewports that is not
    #      active because the $MAXACTVP count is currently being exceeded.
    #  0 = Off
    # <positive value> = On and active. The value indicates the order of
    # stacking for the viewports, where 1 is the active viewport, 2 is the next,
    # and so on:
    'status': DXFAttr(68, default=0),
    'id': DXFAttr(69, default=2),

    # DXF reference: View center point (in DCS):
    # Correction to the DXF reference:
    # This point represents the center point in model space (WCS) stored as
    # 2D point!
    'view_center_point': DXFAttr(12, xtype=XType.point2d, default=NULLVEC),
    'snap_base_point': DXFAttr(13, xtype=XType.point2d, default=NULLVEC),
    'snap_spacing': DXFAttr(14, xtype=XType.point2d, default=Vec3(10, 10)),
    'grid_spacing': DXFAttr(15, xtype=XType.point2d, default=Vec3(10, 10)),
    # View direction vector (WCS):
    'view_direction_vector': DXFAttr(16, xtype=XType.point3d, default=NULLVEC),
    # View target point (in WCS):
    'view_target_point': DXFAttr(17, xtype=XType.point3d, default=NULLVEC),
    'perspective_lens_length': DXFAttr(42, default=50),
    'front_clip_plane_z_value': DXFAttr(43, default=0),
    'back_clip_plane_z_value': DXFAttr(44, default=0),
    # View height (in model space units):
    'view_height': DXFAttr(45, default=1),
    'snap_angle': DXFAttr(50, default=0),
    'view_twist_angle': DXFAttr(51, default=0),
    'circle_zoom': DXFAttr(72, default=100),

    # 331: Frozen layer object ID/handle (multiple entries may exist) (optional)

    # Viewport status bit-coded flags:
    # 1 (0x1) = Enables perspective mode
    # 2 (0x2) = Enables front clipping
    # 4 (0x4) = Enables back clipping
    # 8 (0x8) = Enables UCS follow
    # 16 (0x10) = Enables front clip not at eye
    # 32 (0x20) = Enables UCS icon visibility
    # 64 (0x40) = Enables UCS icon at origin
    # 128 (0x80) = Enables fast zoom
    # 256 (0x100) = Enables snap mode
    # 512 (0x200) = Enables grid mode
    # 1024 (0x400) = Enables isometric snap style
    # 2048 (0x800) = Enables hide plot mode
    # 4096 (0x1000) = kIsoPairTop. If set and kIsoPairRight is not set, then
    #   isopair top is enabled. If both kIsoPairTop and kIsoPairRight are set,
    #   then isopair left is enabled
    # 8192 (0x2000) = kIsoPairRight. If set and kIsoPairTop is not set, then
    #   isopair right is enabled
    # 16384 (0x4000) = Enables viewport zoom locking
    # 32768 (0x8000) = Currently always enabled
    # 65536 (0x10000) = Enables non-rectangular clipping
    # 131072 (0x20000) = Turns the viewport off
    # 262144 (0x40000) = Enables the display of the grid beyond the drawing
    #   limits
    # 524288 (0x80000) = Enable adaptive grid display
    # 1048576 (0x100000) = Enables subdivision of the grid below the set grid
    #   spacing when the grid display is adaptive
    # 2097152 (0x200000) = Enables grid follows workplane switching
    'flags': DXFAttr(90, default=0),

    'clipping_boundary_handle': DXFAttr(340, default=0, optional=True),

    # Plot style sheet name assigned to this viewport
    'plot_style_name': DXFAttr(1, default=''),

    # Render mode:
    # 0 = 2D Optimized (classic 2D)
    # 1 = Wireframe
    # 2 = Hidden line
    # 3 = Flat shaded
    # 4 = Gouraud shaded
    # 5 = Flat shaded with wireframe
    # 6 = Gouraud shaded with wireframe
    'render_mode': DXFAttr(
        281, default=0,
        validator=validator.is_in_integer_range(0, 7),
        fixer=RETURN_DEFAULT,
    ),
    'ucs_per_viewport': DXFAttr(
        71, default=0,
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),
    'ucs_icon': DXFAttr(74, default=0),
    'ucs_origin': DXFAttr(110, xtype=XType.point3d, default=NULLVEC),
    'ucs_x_axis': DXFAttr(
        111, xtype=XType.point3d, default=X_AXIS,
        validator=validator.is_not_null_vector,
        fixer=RETURN_DEFAULT,
    ),
    'ucs_y_axis': DXFAttr(
        112, xtype=XType.point3d, default=Y_AXIS,
        validator=validator.is_not_null_vector,
        fixer=RETURN_DEFAULT,
    ),

    # Handle of AcDbUCSTableRecord if UCS is a named UCS.
    # If not present, then UCS is unnamed:
    'ucs_handle': DXFAttr(345),

    # Handle of AcDbUCSTableRecord of base UCS if UCS is orthographic (79 code
    # is non-zero). If not present and 79 code is non-zero, then base UCS is
    # taken to be WORLD:
    'ucs_base_handle': DXFAttr(346, optional=True),

    # UCS ortho type:
    # 0 = not orthographic
    # 1 = Top
    # 2 = Bottom
    # 3 = Front
    # 4 = Back
    # 5 = Left
    # 6 = Right
    'ucs_ortho_type': DXFAttr(
        79, default=0,
        validator=validator.is_in_integer_range(0, 7),
        fixer=RETURN_DEFAULT,
    ),
    'elevation': DXFAttr(146, default=0),

    # Shade plot mode:
    # 0 = As Displayed
    # 1 = Wireframe
    # 2 = Hidden
    # 3 = Rendered
    'shade_plot_mode': DXFAttr(
        170, dxfversion='AC1018',
        validator=validator.is_in_integer_range(0, 4),
        fixer=RETURN_DEFAULT,
    ),
    # Frequency of major grid lines compared to minor grid lines
    'grid_frequency': DXFAttr(61, dxfversion='AC1021'),
    'background_handle': DXFAttr(332, dxfversion='AC1021'),
    'shade_plot_handle': DXFAttr(333, dxfversion='AC1021'),
    'visual_style_handle': DXFAttr(348, dxfversion='AC1021'),
    'default_lighting_flag': DXFAttr(292, dxfversion='AC1021', default=1,
                                     optional=True),
    # Default lighting type:
    # 0 = One distant light
    # 1 = Two distant lights
    'default_lighting_type': DXFAttr(
        282, default=0, dxfversion='AC1021',
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),
    'view_brightness': DXFAttr(141, dxfversion='AC1021'),
    'view_contrast': DXFAttr(142, dxfversion='AC1021'),

    # as AutoCAD Color Index
    'ambient_light_color_1': DXFAttr(
        63, dxfversion='AC1021',
        validator=validator.is_valid_aci_color,
    ),
    # as True Color:
    'ambient_light_color_2': DXFAttr(421, dxfversion='AC1021'),

    # as True Color:
    'ambient_light_color_3': DXFAttr(431, dxfversion='AC1021'),

    'sun_handle': DXFAttr(361, dxfversion='AC1021'),
    'ref_vp_object_1': DXFAttr(335, dxfversion='AC1021'),
    # unknown meaning, don't ask mozman
    'ref_vp_object_2': DXFAttr(343, dxfversion='AC1021'),
    # unknown meaning, don't ask mozman
    'ref_vp_object_3': DXFAttr(344, dxfversion='AC1021'),
    # unknown meaning, don't ask mozman
    'ref_vp_object_4': DXFAttr(91, dxfversion='AC1021'),
    # unknown meaning, don't ask mozman
})
acdb_viewport_group_codes = group_code_mapping(acdb_viewport)
# Note:
# The ZOOM XP factor is calculated with the following formula:
# group_41 / group_45 (or pspace_height / mspace_height).

FROZEN_LAYER_GROUP_CODE = 331


@register_entity
class Viewport(DXFGraphic):
    """ DXF VIEWPORT entity """
    DXFTYPE = 'VIEWPORT'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_viewport)
    viewport_id = 2

    # Notes to vieport_id:
    # The id of the first viewport has to be 1, which is the definition of
    # paper space. For the following viewports it seems only important, that
    # the id is greater than 1.

    def get_next_viewport_id(self) -> int:
        current_id = Viewport.viewport_id
        Viewport.viewport_id += 1
        return current_id

    def __init__(self):
        super().__init__()
        self._frozen_layers: List[str] = []

    def _copy_data(self, entity: 'Viewport') -> None:
        entity._frozen_layers = self._frozen_layers

    @property
    def frozen_layers(self) -> List[str]:
        """ Set/get frozen layers as list of layer names. """
        return self._frozen_layers

    @frozen_layers.setter
    def frozen_layers(self, names: Iterable[str]):
        self._frozen_layers = list(names)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.fast_load_dxfattribs(
                dxf, acdb_viewport_group_codes, subclass=2, log=False)
            if processor.r12:
                self.load_xdata_into_dxf_namespace()
            else:
                if len(tags):
                    tags = self.load_frozen_layer_handles(tags)
                if len(tags):
                    processor.log_unprocessed_tags(
                        tags, subclass=acdb_viewport.name)
        return dxf

    def post_load_hook(self, doc: 'Drawing'):
        super().post_load_hook(doc)
        bag: List[str] = []
        db = doc.entitydb
        for handle in self._frozen_layers:
            try:
                bag.append(db[handle].dxf.name)
            except KeyError:  # ignore non-existing layers
                pass
        self._frozen_layers = bag

    def load_frozen_layer_handles(self, tags: Tags) -> Tags:
        unprocessed_tags = Tags()
        for tag in tags:
            if tag.code == FROZEN_LAYER_GROUP_CODE:
                self._frozen_layers.append(tag.value)
            else:
                unprocessed_tags.append(tag)
        return unprocessed_tags

    def load_xdata_into_dxf_namespace(self) -> None:
        try:
            tags = [v for c, v in self.xdata.get_xlist('ACAD', 'MVIEW')]
        except DXFValueError:
            return
        tags = tags[3:-2]
        dxf = self.dxf
        flags = 0
        flags = set_flag_state(flags, const.VSF_FAST_ZOOM, bool(tags[11]))
        flags = set_flag_state(flags, const.VSF_SNAP_MODE, bool(tags[13]))
        flags = set_flag_state(flags, const.VSF_GRID_MODE, bool(tags[14]))
        flags = set_flag_state(flags, const.VSF_ISOMETRIC_SNAP_STYLE,
                               bool(tags[15]))
        flags = set_flag_state(flags, const.VSF_HIDE_PLOT_MODE, bool(tags[24]))
        try:
            dxf.view_target_point = tags[0]
            dxf.view_direction_vector = tags[1]
            dxf.view_twist_angle = tags[2]
            dxf.view_height = tags[3]
            dxf.view_center_point = tags[4], tags[5]
            dxf.perspective_lens_length = tags[6]
            dxf.front_clip_plane_z_value = tags[7]
            dxf.back_clip_plane_z_value = tags[8]
            dxf.render_mode = tags[9]  # view_mode == render_mode ?
            dxf.circle_zoom = tags[10]
            # fast zoom flag : tag[11]
            dxf.ucs_icon = tags[12]
            # snap mode flag  = tags[13]
            # grid mode flag = tags[14]
            # isometric snap style = tags[15]
            # dxf.snap_isopair = tags[16] ???
            dxf.snap_angle = tags[17]
            dxf.snap_base_point = tags[18], tags[19]
            dxf.snap_spacing = tags[20], tags[21]
            dxf.grid_spacing = tags[22], tags[23]
            # hide plot flag  = tags[24]
        except IndexError:  # internal exception
            raise DXFStructureError("Invalid viewport entity - missing data")
        dxf.flags = flags
        self._frozen_layers = tags[26:]
        self.xdata.discard('ACAD')

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)
        if tagwriter.dxfversion == DXF12:
            self.export_acdb_viewport_r12(tagwriter)
        else:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_viewport.name)
            self.dxf.export_dxf_attribs(tagwriter, [
                'center', 'width', 'height', 'status', 'id',
                'view_center_point', 'snap_base_point', 'snap_spacing',
                'grid_spacing', 'view_direction_vector', 'view_target_point',
                'perspective_lens_length', 'front_clip_plane_z_value',
                'back_clip_plane_z_value', 'view_height', 'snap_angle',
                'view_twist_angle', 'circle_zoom',
            ])
            if len(self.frozen_layers):
                layers = self.doc.layers
                for layer_name in self.frozen_layers:
                    try:
                        layer = layers.get(layer_name)
                    except DXFTableEntryError:
                        pass
                    else:
                        tagwriter.write_tag2(FROZEN_LAYER_GROUP_CODE,
                                             layer.dxf.handle)

            self.dxf.export_dxf_attribs(tagwriter, [
                'flags', 'clipping_boundary_handle', 'plot_style_name',
                'render_mode', 'ucs_per_viewport', 'ucs_icon', 'ucs_origin',
                'ucs_x_axis', 'ucs_y_axis', 'ucs_handle', 'ucs_base_handle',
                'ucs_ortho_type', 'elevation', 'shade_plot_mode',
                'grid_frequency', 'background_handle', 'shade_plot_handle',
                'visual_style_handle', 'default_lighting_flag',
                'default_lighting_type', 'view_brightness', 'view_contrast',
                'ambient_light_color_1', 'ambient_light_color_2',
                'ambient_light_color_3', 'sun_handle', 'ref_vp_object_1',
                'ref_vp_object_2', 'ref_vp_object_3', 'ref_vp_object_4'
            ])

    def export_acdb_viewport_r12(self, tagwriter: 'TagWriter'):
        self.dxf.export_dxf_attribs(tagwriter, [
            'center', 'width', 'height', 'status', 'id',
        ])
        tagwriter.write_tags(self.dxftags())

    def dxftags(self) -> Tags:
        def flag(flag):
            return 1 if self.dxf.flags & flag else 0

        dxf = self.dxf
        tags = [
            DXFTag(1001, 'ACAD'),
            DXFTag(1000, 'MVIEW'),
            DXFTag(1002, '{'),
            DXFTag(1070, 16),  # extended data version, always 16 for R11/12
            DXFVertex(1010, dxf.view_target_point),
            DXFVertex(1010, dxf.view_direction_vector),
            DXFTag(1040, dxf.view_twist_angle),
            DXFTag(1040, dxf.view_height),
            DXFTag(1040, dxf.view_center_point[0]),
            DXFTag(1040, dxf.view_center_point[1], ),
            DXFTag(1040, dxf.perspective_lens_length),
            DXFTag(1040, dxf.front_clip_plane_z_value),
            DXFTag(1040, dxf.back_clip_plane_z_value),
            DXFTag(1070, dxf.render_mode),
            DXFTag(1070, dxf.circle_zoom),
            DXFTag(1070, flag(const.VSF_FAST_ZOOM)),
            DXFTag(1070, dxf.ucs_icon),
            DXFTag(1070, flag(const.VSF_SNAP_MODE)),
            DXFTag(1070, flag(const.VSF_GRID_MODE)),
            DXFTag(1070, flag(const.VSF_ISOMETRIC_SNAP_STYLE)),
            DXFTag(1070, 0),  # snap isopair ???
            DXFTag(1040, dxf.snap_angle),
            DXFTag(1040, dxf.snap_base_point[0]),
            DXFTag(1040, dxf.snap_base_point[1]),
            DXFTag(1040, dxf.snap_spacing[0]),
            DXFTag(1040, dxf.snap_spacing[1]),
            DXFTag(1040, dxf.grid_spacing[0]),
            DXFTag(1040, dxf.grid_spacing[1]),
            DXFTag(1070, flag(const.VSF_HIDE_PLOT_MODE)),
            DXFTag(1002, '{'),  # start frozen layer list
        ]
        tags.extend(
            DXFTag(1003, layer_name) for layer_name in self.frozen_layers)
        tags.extend([
            DXFTag(1002, '}'),  # end of frozen layer list
            DXFTag(1002, '}'),  # MVIEW
        ])
        return Tags(tags)

    def rename_frozen_layer(self, old_name: str, new_name: str) -> None:
        key = self.doc.layers.key
        old_key = key(old_name)
        self.frozen_layers = [
            (name if key(name) != old_key else new_name)
            for name in self.frozen_layers
        ]



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
from typing import Union

ACAD_13 = 'AC1012'
ACAD_14 = 'AC1014'
ACAD_2000 = 'AC1015'
ACAD_2004 = 'AC1018'
ACAD_2007 = 'AC1021'
ACAD_2010 = 'AC1024'
ACAD_2013 = 'AC1027'
ACAD_2018 = 'AC1032'
ACAD_LATEST = ACAD_2018

SUPPORTED_VERSIONS = [ACAD_13, ACAD_14, ACAD_2000]
HEADER_ID = 0
CLASSES_ID = 1
OBJECTS_ID = 2
SENTINEL_SIZE = 16

Bytes = Union[bytes, bytearray, memoryview]


class DwgError(Exception):
    pass


class DwgVersionError(DwgError):
    pass


class DwgCorruptedFileHeader(DwgError):
    pass


class DwgCorruptedClassesSection(DwgError):
    pass


class DwgCorruptedHeaderSection(DwgError):
    pass


class CRCError(DwgError):
    pass



# Copyright (c) 2019-2020, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Tuple
from ezdxf.math import Vec3, Vec2

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex


class BoundingBox:
    """ 3D bounding box.

    Args:
        vertices: iterable of ``(x, y, z)`` tuples or :class:`Vec3` objects

    """

    def __init__(self, vertices: Iterable['Vertex'] = None):
        if vertices is None:
            self.extmax = None
            self.extmin = None
        else:
            self.extmin, self.extmax = extends(vertices)

    @property
    def has_data(self) -> bool:
        """ Returns ``True`` if data is available """
        return self.extmin is not None

    def inside(self, vertex: 'Vertex') -> bool:
        """ Returns ``True`` if `vertex` is inside bounding box. """
        x, y, z = Vec3(vertex).xyz
        xmin, ymin, zmin = self.extmin.xyz
        xmax, ymax, zmax = self.extmax.xyz
        return (xmin <= x <= xmax) and (ymin <= y <= ymax) and (zmin <= z <= zmax)

    def extend(self, vertices: Iterable['Vertex']) -> None:
        """ Extend bounds by `vertices`.

        Args:
            vertices: iterable of ``(x, y, z)`` tuples or :class:`Vec3` objects

        """
        v = list(vertices)
        if self.has_data:
            v.extend([self.extmin, self.extmax])
        self.extmin, self.extmax = extends(v)

    @property
    def size(self) -> Vec3:
        """ Returns size of bounding box. """
        return self.extmax - self.extmin

    @property
    def center(self) -> Vec3:
        """ Returns center of bounding box. """
        return self.extmin.lerp(self.extmax)


def extends(vertices: Iterable['Vertex']) -> Tuple[Vec3, Vec3]:
    minx, miny, minz = None, None, None
    maxx, maxy, maxz = None, None, None
    for v in vertices:
        v = Vec3(v)
        if minx is None:
            minx, miny, minz = v.xyz
            maxx, maxy, maxz = v.xyz
        else:
            x, y, z = v.xyz
            if x < minx:
                minx = x
            elif x > maxx:
                maxx = x
            if y < miny:
                miny = y
            elif y > maxy:
                maxy = y
            if z < minz:
                minz = z
            elif z > maxz:
                maxz = z
    if minx is None:
        raise ValueError("No vertices give.")
    return Vec3(minx, miny, minz), Vec3(maxx, maxy, maxz)


class BoundingBox2d:
    """ Optimized 2D bounding box.

    Args:
        vertices: iterable of ``(x, y[, z])`` tuples or :class:`Vec3` objects

    """

    def __init__(self, vertices: Iterable['Vertex'] = None):
        if vertices is None:
            self.extmax = None
            self.extmin = None
        else:
            self.extmin, self.extmax = extends2d(vertices)

    @property
    def has_data(self) -> bool:
        """ Returns ``True`` if data is available """
        return self.extmin is not None

    def inside(self, vertex: 'Vertex') -> bool:
        """ Returns ``True`` if `vertex` is inside bounding box. """
        v = Vec2(vertex)
        min_ = self.extmin
        max_ = self.extmax
        return (min_.x <= v.x <= max_.x) and (min_.y <= v.y <= max_.y)

    def extend(self, vertices: Iterable['Vertex']) -> None:
        """ Extend bounds by `vertices`.

        Args:
            vertices: iterable of ``(x, y[, z])`` tuples or :class:`Vec3` objects

        """
        v = list(vertices)
        if self.has_data:
            v.extend([self.extmin, self.extmax])
        self.extmin, self.extmax = extends2d(v)

    @property
    def size(self) -> Vec2:
        """ Returns size of bounding box. """
        return self.extmax - self.extmin

    @property
    def center(self) -> Vec2:
        """ Returns center of bounding box. """
        return self.extmin.lerp(self.extmax)


def extends2d(vertices: Iterable['Vertex']) -> Tuple[Vec2, Vec2]:
    minx, miny = None, None
    maxx, maxy = None, None
    for v in vertices:
        v = Vec2(v)
        x, y = v.x, v.y
        if minx is None:
            minx = x
            maxx = x
            miny = y
            maxy = y
        else:
            if x < minx:
                minx = x
            elif x > maxx:
                maxx = x
            if y < miny:
                miny = y
            elif y > maxy:
                maxy = y
    if minx is None:
        raise ValueError("No vertices give.")
    return Vec2(minx, miny), Vec2(maxx, maxy)



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import (
    SUBCLASS_MARKER, DXF12, MODEL_SPACE_R12, PAPER_SPACE_R12, MODEL_SPACE_R2000,
    PAPER_SPACE_R2000,
)
from ezdxf.math import NULLVEC
from .dxfentity import base_class, SubclassProcessor, DXFEntity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ['Block', 'EndBlk']

acdb_entity = DefSubclass('AcDbEntity', {
    # No auto fix for invalid layer names!
    'layer': DXFAttr(8, default='0', validator=validator.is_valid_layer_name),
    'paperspace': DXFAttr(
        67, default=0, optional=True,
        validator=validator.is_integer_bool,
        fixer=RETURN_DEFAULT,
    ),

})
acdb_entity_group_codes = group_code_mapping(acdb_entity)

acdb_block_begin = DefSubclass('AcDbBlockBegin', {
    'name': DXFAttr(2, validator=validator.is_valid_block_name),
    # The 2nd name with group code 3 is handles internally, and is not an
    # explict DXF attribute.
    'description': DXFAttr(4, default='', optional=True),

    # Flags:
    # 0 = Indicates none of the following flags apply
    # 1 = This is an anonymous block generated by hatching, associative
    #     dimensioning, other internal operations, or an application
    # 2 = This block has non-constant attribute definitions (this bit is not set
    #     if the block has any attribute definitions that are constant, or has
    #     no attribute definitions at all)
    # 4 = This block is an external reference (xref)
    # 8 = This block is an xref overlay
    # 16 = This block is externally dependent
    # 32 = This is a resolved external reference, or dependent of an external
    #      reference (ignored on input)
    # 64 = This definition is a referenced external reference (ignored on input)
    'flags': DXFAttr(70, default=0),
    'base_point': DXFAttr(10, xtype=XType.any_point, default=NULLVEC),
    'xref_path': DXFAttr(1, default=''),
})
acdb_block_begin_group_codes = group_code_mapping(acdb_block_begin)

MODEL_SPACE_R2000_LOWER = MODEL_SPACE_R2000.lower()
MODEL_SPACE_R12_LOWER = MODEL_SPACE_R12.lower()
PAPER_SPACE_R2000_LOWER = PAPER_SPACE_R2000.lower()
PAPER_SPACE_R12_LOWER = PAPER_SPACE_R12.lower()


@register_entity
class Block(DXFEntity):
    """ DXF BLOCK entity """
    DXFTYPE = 'BLOCK'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_block_begin)

    # Block entity flags:
    # This is an anonymous block generated by hatching, associative
    # dimensioning, other internal operations, or an application:
    ANONYMOUS = 1

    # This block has non-constant attribute definitions (this bit is not set
    # if the block has any attribute definitions that are constant, or has no
    # attribute definitions at all):
    NON_CONSTANT_ATTRIBUTES = 2

    # This block is an external reference:
    XREF = 4

    # This block is an xref overlay:
    XREF_OVERLAY = 8

    # This block is externally dependent:
    EXTERNAL = 16

    # This is a resolved external reference, or dependent of an external reference:
    RESOLVED = 32

    # This definition is a referenced external reference:
    REFERENCED = 64

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor is None:
            return dxf
        processor.fast_load_dxfattribs(dxf, acdb_entity_group_codes, 1)
        processor.fast_load_dxfattribs(dxf, acdb_block_begin_group_codes, 2)
        if processor.r12:
            name = dxf.name.lower()
            if name == MODEL_SPACE_R12_LOWER:
                dxf.name = MODEL_SPACE_R2000
            elif name == PAPER_SPACE_R12_LOWER:
                dxf.name = PAPER_SPACE_R2000
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)

        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_entity.name)
        if self.dxf.hasattr('paperspace'):
            tagwriter.write_tag2(67, 1)
        self.dxf.export_dxf_attribs(tagwriter, 'layer')
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_block_begin.name)

        name = self.dxf.name
        if tagwriter.dxfversion == DXF12:
            # export modelspace and paperspace with leading '$' instead of '*'
            if name.lower() == MODEL_SPACE_R2000_LOWER:
                name = MODEL_SPACE_R12
            elif name.lower() == PAPER_SPACE_R2000_LOWER:
                name = PAPER_SPACE_R12

        tagwriter.write_tag2(2, name)
        self.dxf.export_dxf_attribs(tagwriter, ['flags', 'base_point'])
        tagwriter.write_tag2(3, name)
        self.dxf.export_dxf_attribs(tagwriter, ['xref_path', 'description'])

    @property
    def is_layout_block(self) -> bool:
        """ Returns ``True`` if this is a :class:`~ezdxf.layouts.Modelspace` or
        :class:`~ezdxf.layouts.Paperspace` block definition.
        """
        name = self.dxf.name.lower()
        return name.startswith('*model_space') or name.startswith(
            '*paper_space')

    @property
    def is_anonymous(self) -> bool:
        """ Returns ``True`` if this is an anonymous block generated by
        hatching, associative dimensioning, other internal operations, or an
        application.

        """
        return self.get_flag_state(Block.ANONYMOUS)

    @property
    def is_xref(self) -> bool:
        """ Returns ``True`` if bock is an external referenced file."""
        return self.get_flag_state(Block.XREF)

    @property
    def is_xref_overlay(self) -> bool:
        """ Returns ``True`` if bock is an external referenced overlay file. """
        return self.get_flag_state(Block.XREF_OVERLAY)


acdb_block_end = DefSubclass('AcDbBlockEnd', {})


@register_entity
class EndBlk(DXFEntity):
    """ DXF ENDBLK entity """
    DXFTYPE = 'ENDBLK'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_block_end)

    def load_dxf_attribs(self,
                         processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_entity_group_codes, 1)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)

        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_entity.name)
        if self.dxf.hasattr('paperspace'):
            tagwriter.write_tag2(67, 1)
        self.dxf.export_dxf_attribs(tagwriter, 'layer')
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_block_end.name)
